% Written by Anne Gehre
% Implemenation of "Interactive Curve Constrained Functional Maps", by
% Anne Gehre, Michael Bronstein, Leif Kobbelt, and Justin Solomon (SGP
% 2018)

clear all; close all; 
addpath(genpath(pwd));

%% Load meshes and compute Laplacian eigendecomposition
global n_samples_fourier  
global numEigsTar numEigsSrc OpTar OpSrc

% Number of basis vectors for computing the functional map.
numEigsSrc = 120;
numEigsTar = 120;
landmarks = [];
texture_im = 'Data/texture.jpg';
im = imread(texture_im);


file1 = 'Data/dog2';
file2 =  'Data/pig_388';
file1_curves = strcat(file1, '_curves_dog2_pig_388.txt');
file2_curves = strcat(file2 , '_curves_dog2_pig_388.txt');

[X, T] = readOff(file1);
fprintf('reading the source shape...');tic;
Src = MeshInfoSimple(X, T, numEigsSrc);
fprintf('done (found %d vertices)\n',Src.nv);toc;

fprintf('reading the target shape...');tic;
[X, T] = readOff(file2);
Tar = MeshInfoSimple(X, T, numEigsTar);
fprintf('done (found %d vertices)\n',Tar.nv);toc;

%% load curves 
ncurves = 0;
try
    curveodering_src = importdata(file1_curves);
    ncurves = max(curveodering_src(:,1));
catch
    curveodering_src = [];
end
try
    curveodering_tar = importdata(file2_curves);
catch
    curveodering_tar = [];
end

curvemap = [];
if ncurves>0
    curvemap = repmat(linspace(1,ncurves,ncurves)', 1,2);
end

SrcLaplaceBasis = Src.laplaceBasis; SrcEigenvalues = Src.eigenvalues;
TarLaplaceBasis = Tar.laplaceBasis; TarEigenvalues = Tar.eigenvalues;
Src.laplaceBasis = SrcLaplaceBasis(:,1:numEigsSrc); Src.eigenvalues = SrcEigenvalues(1:numEigsSrc);
Tar.laplaceBasis = TarLaplaceBasis(:,1:numEigsTar); Tar.eigenvalues = TarEigenvalues(1:numEigsTar);

%% Optional: add Descriptors
fct_src = [];
load_shot = 0;
if load_shot
    fprintf('Computing the descriptors...\n');tic;
    try
        fct_src = [fct_src, importdata(fileshot_src)];
    catch
        fct_src = ones(size(Src.X,1),1);
    end
    %fct_src = [fct_src, waveKernelMap(SrcLaplaceBasis, SrcEigenvalues, Src.Ae, 200, landmarks(:,1))];

    fct_tar = [];
    try
        fct_tar = [fct_tar, importdata(fileshot_tar)];
    catch
        fct_tar = ones(size(Tar.X,1),1);
    end
    %fct_tar = [fct_tar, waveKernelMap(TarLaplaceBasis, TarEigenvalues, Tar.Ae, 200, landmarks(:,2))];

    fprintf('done computing descriptors (%d on source and %d on target)\n',size(fct_src,2),size(fct_tar,2)); toc;

    assert(size(fct_src,2)==size(fct_tar,2));
else
    fct_src = ones(size(Src.X,1),1);
    fct_tar = ones(size(Tar.X,1),1);
end
 
%% Prepare Curve Objective Function
n_samples_fourier = 50;

Curves_src = [];
Curves_tar = [];
Curves_src_cell = {};
Curves_tar_cell = {};
c=0;
Hs_tarc = {};
Hs_srcc = {};

for i=1:size(curvemap,1)
    k = find(curveodering_src(:,1)==curvemap(i,1));
    if ~isempty(k)
        c = c+1;
        [~, ~, P_src, curve_vids_src] = fourierBasis(Src, curveodering_src, curvemap(i,1), fct_src,n_samples_fourier, false);
        H_src = P_src * Src.laplaceBasis;
        Hs_srcc{end+1} = H_src;       
        Curves_src = [Curves_src; curve_vids_src];
        Curves_src_cell{end+1} = curve_vids_src;

        [~, ~, P_tar, curve_vids_tar] = fourierBasis(Tar, curveodering_tar, curvemap(i,2), fct_tar,n_samples_fourier, false);
        H_tar = P_tar * Tar.laplaceBasis;
        Hs_tarc{end+1} = H_tar;
        Curves_tar = [Curves_tar; curve_vids_tar];
        Curves_tar_cell{end + 1} = curve_vids_tar;
    end
end

%% Optional: Add point correspondences for wave kernel map
constFct = sign(Src.laplaceBasis(1,1)*Tar.laplaceBasis(1,1))*[sqrt(sum(Tar.area)/sum(Src.area)); zeros(numEigsTar-1,1)];
funProj = @(F) [constFct; F(numEigsTar+1:end)];
F_lb = zeros(numEigsTar*numEigsSrc, 1); F_lb(1) = constFct(1);
F_lb = reshape(F_lb, numEigsTar, numEigsSrc);
% fig = figure;
% pF_lb = ones(size(Tar.X,1),1);
% fig = showFunctionalMap(Src, Tar, Curves_src_cell, Curves_tar_cell, F_lb, pF_lb, 1, im, fig);
% x = 'y';
% while x~='n' & x~='N'
%     prompt = 'Would you like to add another point correspondence [y/n]?';
%     x = input(prompt, 's');
%     if x~='n' & x~='N'
%         fig = showFunctionalMap(Src, Tar, Curves_src_cell, Curves_tar_cell, F_lb, pF_lb, 1, S, fig);
%         figure(fig);
%         cameratoolbar;
%         idx0 = selectPoint(fig, Src);
%         idx1 = selectPoint(fig, Tar);
%         landmarks = [landmarks; idx0,idx1];
%     end
% end
% if size(landmarks,2)>0
%     fct_src = [fct_src, waveKernelMap(SrcLaplaceBasis, SrcEigenvalues, Src.Ae, 200, landmarks(:,1))];
%     fct_tar = [fct_tar, waveKernelMap(TarLaplaceBasis, TarEigenvalues, Tar.Ae, 200, landmarks(:,2))];
% end
% 
% % Normalization
% no = sqrt(diag(fct_src'*Src.Ae*fct_src))';
% fct_src = fct_src ./ repmat(no, [Src.nv,1]);
% no = sqrt(diag(fct_tar'*Tar.Ae*fct_tar))';
% fct_tar = fct_tar ./ repmat(no, [Tar.nv,1]);
% 
% 
% fprintf('Pre-computing the multiplication operators...');tic;
% %% Multiplication Operators
% numFct = size(fct_src,2);
% OpSrc = cell(numFct,1);
% OpTar = cell(numFct,1);
% for i = 1:numFct
%     OpSrc{i} = Src.laplaceBasis'*Src.Ae*(repmat(fct_src(:,i), [1,numEigsSrc]).*Src.laplaceBasis);
%     OpTar{i} = Tar.laplaceBasis'*Tar.Ae*(repmat(fct_tar(:,i), [1,numEigsTar]).*Tar.laplaceBasis);
% end
% 
% Fct_src = Src.laplaceBasis'*Src.Ae*fct_src;
% Fct_tar = Tar.laplaceBasis'*Tar.Ae*fct_tar;
%fprintf('done\n');toc;


%% Optimize the FM
a = 0;       % Descriptors preservation
b = 0;       % Commutativity with descriptors
c = 0.05;    % Commutativity with Laplacian 
d = 1;       % Feature Curve Constraints
nconstraints = 0;%ceil(length(OpSrc)); %number of Descriptor preservation constraints

ATAs = zeros(numEigsSrc, numEigsTar);
As = [];
Bs = [];
ATs = [];
BTs = [];
BBTs =  zeros(numEigsSrc, numEigsTar);
ETDs =  zeros(numEigsSrc, numEigsTar);
ETEs =  zeros(numEigsSrc, numEigsTar);

%add laplace constraints
B = full(spdiags(Src.eigenvalues, 0 , length(Src.eigenvalues), length(Src.eigenvalues)));
A = full(spdiags(Tar.eigenvalues, 0 , length(Tar.eigenvalues), length(Tar.eigenvalues)));
ATAs = ATAs +c*A'*(A); 
As=[As; c*A]; 
Bs=[Bs;B]; 
ATs=[ATs;c*A']; 
BTs=[BTs;B']; 
BBTs = BBTs +c* B*(B');

%add permutation constraints
for i = 1:nconstraints
    A = full(OpTar{i});
    B = full(OpSrc{i});
    ATAs = ATAs + b*A'*(A); 
    As=[As; b*A]; 
    Bs=[Bs;B]; 
    ATs=[ATs;b*A']; 
    BTs=[BTs;B']; 
    BBTs =BBTs + b* B*(B');
end
 
%curve constraints
ncurves = length(Hs_srcc);
for i = 1:ncurves
    D = Hs_srcc{i};
    E = Hs_tarc{i};
    ETEs = ETEs + d*E'*E;
    ETDs = ETDs + d*E'*D;
end
       

%precompute index matrix for conjgrad
colsA = size(As,1);
dimC = numEigsSrc;
idx = repmat(linspace(1, dimC, dimC)', 1 , colsA);
idx = idx + repmat((nconstraints+1)*dimC*linspace(0, dimC-1, dimC), dimC, (nconstraints+1));

for i=1:nconstraints
    idx(:, i*dimC+1:(i+1)*dimC) = idx(:, i*dimC+1:(i+1)*dimC ) + i*dimC;
end
    

%% interactive map compuation
x = 'y';
Yg = Tar.laplaceBasis;
Xg = Src.laplaceBasis;

fig = figure;
while x~='n' & x~='N'
    % Compute Functional map with conjugate gradients
    tic;
    F_lb = conjgrad(ATAs + ETEs,As,Bs,b,ATs, BTs, BBTs,ETDs,F_lb, nconstraints+1, length(Hs_srcc), idx);
    toc;
    % Compute p2p correspondences according to Ovsjanikov 2012
%     tic; 
%     [dist, pF_lb ]= knn2((F_lb*Xg')',Yg);
%     toc;

    %map visualization
    fig = showFunctionalMap(Src, Tar, Curves_src_cell, Curves_tar_cell, F_lb, 1, im, fig);
    [ncurves x curveodering_src curveodering_tar c1 c2] = addcurve(Src, Tar, curveodering_src, curveodering_tar, fig, ncurves);
    
    % update curve constraints
    [~, ~, P_src, curve_vids_src] = fourierBasis(Src, curveodering_src, ncurves, fct_src,n_samples_fourier, false);
    H_src = P_src * Src.laplaceBasis;
    Hs_srcc{end+1} = H_src;       
    Curves_src = [Curves_src; curve_vids_src];
    Curves_src_cell{end+1} = curve_vids_src;

    [~, ~, P_tar, curve_vids_tar] = fourierBasis(Tar, curveodering_tar, ncurves, fct_tar,n_samples_fourier, false);
    H_tar = P_tar * Tar.laplaceBasis;
    Hs_tarc{end+1} = H_tar;
    Curves_tar = [Curves_tar; curve_vids_tar];
    Curves_tar_cell{end + 1} = curve_vids_tar;
    
    D = Hs_srcc{end};
    E = Hs_tarc{end};
    ETEs = ETEs + d*(E')*E;
    ETDs = ETDs + d*E'*D;
end
    fprintf('done fmap optimization. Store curve Files. \n');
    %store the curves
    fileID = fopen(file1_curves,'w');
    fprintf(fileID,'%d %d\n', curveodering_src');
    fclose(fileID);

    fileID = fopen(file2_curves,'w');
    fprintf(fileID,'%d %d\n', curveodering_tar');
    fclose(fileID);

%% Evaluation
% Compute the p2p map
fprintf('Visualizing results (without ICP) \n');
tic; 
[dist, pF_lb ]= knn2((F_lb*Xg')',Yg);
toc;


figure;
nbasis = size(Src.laplaceBasis,2);
kf = rand(nbasis,1)-0.5;
kf = kf.*((nbasis:-1:1).^2');

f = Src.laplaceBasis*kf;
F = Src.laplaceBasis'*Src.Ae*f;
g = Tar.laplaceBasis*F_lb*F;
f = Src.laplaceBasis*F;
subplot(1,2,1);
trimesh(Src.T, Src.X(:,1), Src.X(:,2), Src.X(:,3),f,'FaceColor','interp');
plotcurves(Curves_src_cell, Src.X);
view([0 90]);
title('f = random smooth function');
axis equal;
axis off;
caxis manual;
caxis([min(f) max(f)]);
colorbar;


subplot(1,2,2);
trimesh(Tar.T, Tar.X(:,1), Tar.X(:,2), Tar.X(:,3),g,'FaceColor','interp');
hold on;
plotcurves(Curves_tar_cell, Tar.X);
view([0 90]);
title('f(p2pmap) before ICP');
axis equal;
axis off;
axis off;
caxis manual;
caxis([min(f) max(f)]);
colorbar;


%%
figure;
%samples = euclidean_fps(Src,200);
samples = Curves_src;
for i=1:length(samples)
    if samples(i)>size(pF_lb)
        samples(i)=1;
    end
end
visualize_map_lines(Tar, Src, pF_lb, samples);
plotcurves(Curves_tar_cell, Tar.X);
points = Src.X;
xdiam = 3/2*(max(points(:,1))-min(points(:,1)));
points(:,1) = points(:,1) + xdiam;
plotcurves(Curves_src_cell, points);

view([0 90]);
title('P2P map (Ovsjanikov 2012)');
axis equal;
axis off;


